/*
 *  This sample code is public domain. Use it however you wish at your own risk. 
 */ 

package org.example.georg.connector.gc;

import android.content.ContentResolver;
import android.content.ContentValues;
import android.content.Context;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.pm.PackageManager.NameNotFoundException;
import android.net.Uri;
import android.os.Bundle;
import android.preference.EditTextPreference;
import android.preference.PreferenceActivity;
import android.preference.PreferenceManager;
import android.preference.PreferenceScreen;
import android.view.inputmethod.EditorInfo;
import android.widget.Toast;

/**
 * Small preference-layout to customize the connector.
 * 
 *
 */
public class GCHttpConnectorSettings extends PreferenceActivity {

	/** Called when the activity is first created. */
	public static final String SETTINGS_USER = "settings_user";
	public static final String SETTINGS_PASSWORD = "settings_password";
	public static final String SETTINGS_HTTP_TIMEOUT = "settings_http_timeout";
	public static final String SETTINGS_ID = "GeorgGCConn";
	public static final String SETTINGS_DEFAULT_NETWORK_TIMEOUT = "20";
	
	/**
	 * So that we can notify the Loader in case settings are changed
	 */
	private static GCLoader cvGCLoader;

	/**
	 * So that we can notify the Loader in case settings are changed
	 */
	public static void registerGCLoader(GCLoader aLoader) {
		cvGCLoader = aLoader;
	}
	
	/**
	 * reset the Loader
	 */
	public static void resetGCLoader() {
		if (null != cvGCLoader) {
			cvGCLoader.reset();
		}
	}
	
	public static String getVersion(Context aCtx) {
		try {
			PackageInfo info = aCtx.getPackageManager().getPackageInfo(aCtx.getPackageName(), PackageManager.GET_META_DATA);
			return info.versionName;
		} catch (NameNotFoundException e) {
			return "unknown";
		}
	}

	@Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
       
        // Load preference-Layout from ./res/xml/gccon_settings.xml
		setPreferenceScreen(createPreferenceHierarchy());
		setTitle("GC.com-Connector " + getVersion(this));
		
    }
    
    @Override
    protected void onResume() {
    	super.onResume();
    	// Whenever we're started, we register with GeOrg
		registerConnector();
    }
    
    @Override
    protected void onPause() {
    	super.onPause();
    	// After having visited the settings, tell the loader to reset
    	resetGCLoader();
    }
    

    /**
     * Build the layout for our preferences from xml
     * 
     * @return
     */
    private PreferenceScreen createPreferenceHierarchy() {
        // Root
    	PreferenceManager lvManager = getPreferenceManager();
    	lvManager.setSharedPreferencesMode(Context.MODE_PRIVATE);   	
    	lvManager.setSharedPreferencesName(SETTINGS_ID);
        PreferenceScreen root = lvManager.createPreferenceScreen(this);    
      
        // Edit text preference
        EditTextPreference editTextPref = new EditTextPreference(this);
        editTextPref.setDialogTitle("User for geocaching.com");
        editTextPref.setKey(SETTINGS_USER);
        editTextPref.setTitle("User for geocaching.com");
        editTextPref.setSummary("This will be used, when downloading GPX.");
        editTextPref.getEditText().setLines(1);
        editTextPref.getEditText().setInputType(EditorInfo.TYPE_CLASS_TEXT);
        
        root.addPreference(editTextPref);
        
        // Edit text preference
        editTextPref = new EditTextPreference(this);
        editTextPref.setDialogTitle("Password for geocaching.com");
        editTextPref.setKey(SETTINGS_PASSWORD);
        editTextPref.setTitle("Password for geocaching.com");
        editTextPref.setSummary("This will be used, when downloading GPX.");
        editTextPref.getEditText().setLines(1);
        editTextPref.getEditText().setInputType(EditorInfo.TYPE_CLASS_TEXT |EditorInfo.TYPE_TEXT_VARIATION_PASSWORD);
        root.addPreference(editTextPref);
        
        // Edit text preference
        editTextPref = new EditTextPreference(this);
        editTextPref.setDialogTitle("Timeout for Web-Requests");
        editTextPref.setKey(SETTINGS_HTTP_TIMEOUT);
        editTextPref.setTitle("Timeout for Web-Requests");
        editTextPref.setSummary("In seconds. Change only if you know what you're doing.");
        editTextPref.setDefaultValue(SETTINGS_DEFAULT_NETWORK_TIMEOUT);
        editTextPref.getEditText().setLines(1);
        editTextPref.getEditText().setInputType(EditorInfo.TYPE_CLASS_NUMBER);
        root.addPreference(editTextPref);
       
        return root;
    }
    
    /**
     * Register with GeOrg. This is where you have to change name and uri to your own connector
     */
    private void registerConnector() {
		ContentResolver lvContentResolver = getContentResolver();
		ContentValues lvContent = new ContentValues();
		
		// This is displayed in GeOrg
		lvContent.put("georg_connector_name", "Groundspeak Connector");
		
		// This prefix is used when trying to find the connector and must match with the classname
		// in AndroidManifest.xml!!
		lvContent.put("georg_connector_uri", "org.example.georg.connector.gc.GCHttpConnector");

		try {
			// GeOrgs own URL for registration. Don't change this
			lvContentResolver.insert(
				Uri.parse("content://de.ranitos.georg.connector/registerGeorgConnector"), 
				lvContent);
			Toast.makeText(this, "Registered with GeOrg!", Toast.LENGTH_SHORT).show();
		} catch (IllegalArgumentException anExc) {
			Toast.makeText(this, "Could not register with GeOrg - GeOrg not found!", Toast.LENGTH_SHORT).show();			
			return;
		}
    }
}