/*
 *  This sample code is public domain. Use it however you wish at your own risk. 
 */ 

package org.example.georg.connector.gc;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.nio.charset.IllegalCharsetNameException;
import java.nio.charset.UnsupportedCharsetException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.GZIPInputStream;
import java.util.zip.Inflater;
import java.util.zip.InflaterInputStream;

import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.CookieStore;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpUriRequest;
import org.apache.http.client.params.ClientPNames;
import org.apache.http.client.protocol.ClientContext;
import org.apache.http.conn.scheme.PlainSocketFactory;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.cookie.Cookie;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.BasicCookieStore;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.tsccm.ThreadSafeClientConnManager;
import org.apache.http.message.AbstractHttpMessage;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.params.HttpParams;
import org.apache.http.protocol.BasicHttpContext;
import org.apache.http.protocol.HttpContext;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.database.Cursor;
import android.database.MatrixCursor;
import android.net.ConnectivityManager;
import android.util.Log;

/**
 * Sample for a connector-implementation. Just so that our Connector-Sample doesn't do nothing.
 * This code uses the Groundspeak GoogleMaps-Mashup API to find caches and allows for downloading
 * GPX from the site.
 * 
 * It behaves just like a (very specialised) browser and only connects to Groundspeak upon
 * user-interaction and within a normal traffic-range. Thus it can't be called a spider, bot or crawler
 * and so might be ok according to Groundspeaks Terms of Use. There is no official statement
 * from Groundspeak though, so use it at your own risk.
 * 
 * Also the interface used here is not officially documented. It may be subject to sudden
 * change or cease to exist altogether. Don't rely on this code in any way.
 * 
 * HISTORY:
 * 2012.01.01: 1.0.8 Dumped HttpUrlConnection, included HttpClient
 * 2011.10.11: 1.0.7 Fixed bugs at spoiler-download and userdata-handling
 * 2010.05.09: 1.0.6 Fixed some URLs
 * 2010.12.13: 1.0.5 Fixed a bug that wouldn't download GPX when not downloading spoilers
 * 2010.11.13: 1.0.4 Fixed some bugs in pushLog
 * 2010.11.12: 1.0.3 Fixed some bugs in readCachesbyCoord and readCacheDetails
 * 2010.06.25: 1.0.2 Added some more queries
 * 2010.03.04: 1.0.1 Allow for providers to proxy-zip data
 * 2010.02.14: 1.0.0 Allow spoiler-download 
 * 2009.12.28:       Connection Timeout now equals Read-Timeout
 * 2009.10.18:       Small Bugfix
 * 2009.10.10:       Initial Release
 * 
 */

public class GCLoader {
	
	private static final String TAG = "GCLoader";
	
	private static final boolean DEBUG_WAYPOINT = false; // will add a parking position to each cache for testing
	private static final boolean DEBUG_MULTI_GPX_DOWNLOAD = false; // will only point to an existing GPX-file for testing
	private static final boolean DEBUG_SYNC_CACHES = false; // will only point to an existing GPX-file for testing
	
	// Some constant changeable stuff
	private static final String URL_MAP_INIT = "http://www.geocaching.com/map/default.aspx";
	private static final String URL_LOGIN = "https://www.geocaching.com/login/default.aspx?RESETCOMPLETE=Y";
	private static final String URL_LOGIN_GET = "https://www.geocaching.com/login/default.aspx";
	private static final String URL_LISTING = "http://www.geocaching.com/seek/cache_details.aspx?wp=";
	private static final String URL_LOG = "http://www.geocaching.com/seek/log.aspx?wp=";
	private static final String URL_FIELD_NOTES = "http://www.geocaching.com/my/uploadfieldnotes.aspx";
	private static final String GCMAPS_HOST = "www.geocaching.com";
	private static final String SPOILER_AREA_PREFIX = "ctl00_ContentBody_Images";
	private static final String SPOILER_AREA_END_PREFIX = "ctl00_ContentBody_lblFindCounts";
	private static final String SPOILER_PREFIX = "http://img.geocaching.com/cache/";
	private static final String SPOILER_NAME_PREFIX = "<span>";
	private static final String MULTIPART_PREFIX="--";
	//private static final String GCMAPS_HOST = "web06.groundspeak.com";
	
	/** session-state */
	private String ivLastViewState;
	
	/** map-request optimization */
	private boolean ivLastRequestFromMap;
	
	/** Used after errors to re-establish the connection */
	private boolean ivIsAlreadyInited;

	private String ivUser;
	private String ivSession;
	private Context ivCtx;
	
	private String mapUserToken;

	private DefaultHttpClient ivHttpClient;

	private CookieStore ivCookieStore;

	private HttpContext ivLocalContext;


	/**
	 * Just saves the Context parm for future use
	 * @param aCtx
	 */
	public GCLoader(Context aCtx) {
		ivCtx = aCtx;
		GCHttpConnectorSettings.registerGCLoader(this);
	}
	
	/**
	 * Get or create and configure a "singleton" HttpClient
	 * @return
	 */
	private HttpClient getHttpClient() {
		if (null == ivHttpClient) {
			
			HttpParams lvParams = new DefaultHttpClient().getParams();
			lvParams.setParameter(
					CoreConnectionPNames.SO_TIMEOUT, 
					1000*Integer.valueOf(getPreference(
							GCHttpConnectorSettings.SETTINGS_HTTP_TIMEOUT, 
							GCHttpConnectorSettings.SETTINGS_DEFAULT_NETWORK_TIMEOUT)));
			
			lvParams.setParameter(
					CoreConnectionPNames.CONNECTION_TIMEOUT, 
					1000*Integer.valueOf(getPreference(
							GCHttpConnectorSettings.SETTINGS_HTTP_TIMEOUT, 
							GCHttpConnectorSettings.SETTINGS_DEFAULT_NETWORK_TIMEOUT)));
			
			lvParams.setBooleanParameter(ClientPNames.HANDLE_REDIRECTS, false);
			
			SchemeRegistry schemeRegistry = new SchemeRegistry();
			schemeRegistry.register(
			         new Scheme("http", PlainSocketFactory.getSocketFactory(), 80));
			schemeRegistry.register(
			         new Scheme("https", SSLSocketFactory.getSocketFactory(), 443));

			ThreadSafeClientConnManager cm = new ThreadSafeClientConnManager(lvParams, schemeRegistry);
			
			ivHttpClient = new DefaultHttpClient(cm, lvParams);
			ivLocalContext = new BasicHttpContext();
			ivCookieStore = new BasicCookieStore();
			
			// Bind custom cookie store to the local context
			ivLocalContext.setAttribute(ClientContext.COOKIE_STORE, ivCookieStore);
			
			
		}
		return ivHttpClient;
	}
	
	/**
	 * Convenience-Wrapper around executing the actual Http-Request
	 * @param aRequest
	 * @return
	 * @throws ConnectorException
	 */
	private HttpResponse executeHttpRequest(HttpUriRequest aRequest) throws ConnectorException {
		
		connectedOrThrow();
		
		try {
			return getHttpClient().execute(aRequest, ivLocalContext);
		} catch (ClientProtocolException e1) {
			Log.e(TAG, "ClientProtocolException", e1);
			throw new ConnectorException(e1.getMessage());
		} catch (IOException e1) {
			Log.e(TAG, "IOException", e1);
			throw new ConnectorException(e1.getMessage());
		}
	}
		
	/**
	 * Returns a list of all queries, this connector supports
	 * @return
	 */
	public Cursor getSupportedQueries() {
		MatrixCursor lvQueries = new MatrixCursor(new String[] {"query"});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.GET_SUPPORTED_QUERIES});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.PING});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.LIST_CACHES_BY_COORD});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.GET_CACHE_DETAILS});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.GET_CACHE_GPX});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.GET_SPOILERS});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.PUSH_LOG});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.PUSH_FIELD_NOTES});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.GET_COMMANDS});
		lvQueries.addRow(new String[]{ConnectorSupportedQueries.EXECUTE_COMMAND});
		if (DEBUG_MULTI_GPX_DOWNLOAD) {
			lvQueries.addRow(new String[]{ConnectorSupportedQueries.DOWNLOAD_MULTIPLE});
		}
		if (DEBUG_SYNC_CACHES) {
			lvQueries.addRow(new String[]{ConnectorSupportedQueries.SYNC_CACHES});
		}
		return lvQueries;
	}
	
	
	/**
	 * Returns a list of all commands this connector supports
	 * The resulting table has this information as rows:
	 * 
	 * "name"    - internal name of the command
	 * "readable" - user readable representation of the command for display in GeOrg
	 * 

	 * @return
	 */
	public Cursor getCommands() {
		MatrixCursor lvCommands = new MatrixCursor(new String[] {"name", "readable"});
		lvCommands.addRow(new String[]{"settings", "Configure"});
		lvCommands.addRow(new String[]{"version", "Query version"});
		return lvCommands;
	}
	
	/**
	 * Execute one of the commands gathered from getCommands().
	 * Returns a msg. If it's empty, nothing is displayed in GeOrg. If it contains "ok", a
	 * success-Toast is shown in GeOrg. If it contains any other text, this is displayed in a
	 * Dialog-Box in GeOrg.
	 * @return
	 */
	public Cursor executeCommand(String aCmd) {
		MatrixCursor lvCommands = new MatrixCursor(new String[] {"message"});
		if ("settings".equals(aCmd)) {
			Intent lvIntent = new Intent(ivCtx, GCHttpConnectorSettings.class);
			lvIntent.addFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP|Intent.FLAG_ACTIVITY_NEW_TASK);
			ivCtx.startActivity(lvIntent);
		} else if ("version".equals(aCmd)) {
			lvCommands.addRow(new String[]{"GCHttpConnector-sample "+GCHttpConnectorSettings.getVersion(ivCtx)});
		} else {
			lvCommands.addRow(new String[]{"Unknown command "+aCmd});
		}
		
		return lvCommands;
	}

	/**
	 * Read a list of caches within a coordinate rectangle. Returns a Cursor to the Result. The
	 * result may contain caches and additional waypoints
	 * The resulting table has this information as rows:
	 * 
	 * key         as cache                      as waypoint
	 * ----------+-----------------------------+-------------------
	 * "gcid"    - GC code                       WP code
	 * "name"    - cachename                     Waypoint name
	 * "type"    - cachetype                     Waypoint type
	 * "lat"     - latitude of cache             ... of waypoint
	 * "lon"     - longiture of cache            ... of waypoint
	 * "id"      - later used when getting       not used
	 *             the cache-details
	 * "active"  - is cache active?              not used
	 * "found"   - has user found cache?         not used
	 * "asscgc"  - not used                      gcid of associated cache
	 * "asscnm"  - not used                      name of associated cache
	 * "message" - error-message                 error-message
	 * 
	 * Errors are given in the column "message". If it's not null this will be 
	 * interpreted as an error in the frontend.
	 * 
	 * @param aGCID
	 * @return
	 */
	@SuppressWarnings("unused")
	public synchronized Cursor readCachesByCoord( 
			double aTop, 
			double aLeft, 
			double aBottom,
			double aRight,
			String aFilter,
			boolean aShowAddWP) {
	
		MatrixCursor lvContainer = new MatrixCursor(new String[] {"gcid", "name", "type", "lat", "lon", "id", "active", "found", "asscgc", "asscnm", "message"});
	
		Log.v(TAG, "reading Caches by coord");
		
		try {
					
			boolean lvRC = init();
	
			if (!lvRC) {
				lvContainer.addRow(new String[] {null, null, null, null, null, null, null, null, null, null, "Could not login!"});
				return lvContainer;	
			}
			
			// set viewstate to map-page
			if (!ivLastRequestFromMap) {
				String initRequest = initRequest("http://www.geocaching.com/map/default.aspx", false);
				mapUserToken = getPart(initRequest, "userToken = '", "';");
				ivLastRequestFromMap = true;
				
			}
			
			HashMap<String, String> lvFilter = parseCacheFilter(aFilter);
					
			// build the request by coord. mainly copied from network sniffer
			String lvBody = new StringBuilder()
			.append("{\"dto\":{\"data\":{\"c\":1,\"m\":\"\",\"d\":\"")
			.append(aTop).append("|")
			.append(aBottom).append("|")
			.append(aRight).append("|")
					.append(aLeft)
			.append("\"},\"ut\":\"").append(mapUserToken).append("\"}}")
			.toString();
					
			// build the url for the request. the coords don't matter in the url.
			String lvUrlString = "http://"+GCMAPS_HOST+"/map/default.aspx/MapAction";
			URI lvUrl;
			try {
				lvUrl = new URI(lvUrlString);
			} catch (URISyntaxException e) {
				throw new ConnectorException("Could not parse Url "+lvUrlString);
			}
			
			HttpPost post = new HttpPost(lvUrl);
			addHeaders(post, URL_MAP_INIT, true );		
			post.setHeader("Content-Type", "application/json; charset=utf-8");
			post.setHeader("X-Requested-With", "XMLHttpRequest");
			post.setHeader("Accept", "application/json, text/javascript, */*; q=0.01");
			setPostBody(post, lvBody);
			HttpResponse lvResponse = executeHttpRequest(post);
	
			String lvResult = readResponse(lvResponse, false);
			HashMap<String, String> lvGCLookup = new HashMap<String, String>();
			
			// Here comes JSON with the data...
			JSONObject lvObj;
			try {
				lvObj = new JSONObject(lvResult);
				String lvInnerString = (String)lvObj.get("d");
				lvObj = new JSONObject(lvInnerString);
				JSONObject lvCS = lvObj.getJSONObject("cs");
			
			
				// has data?
				if (!lvCS.has("cc")) {
					return lvContainer;
				}
				JSONArray lvCA = lvCS.getJSONArray("cc");
				
				//Log.v(TAG, lvObj.toString(2));
				//Log.v(TAG, "Number Caches: " + lvCS.getInt("count"));
				
				
				// parse result and construct resultset
				int lvCount = lvCS.getInt("count");
				for (int i = 0; i<lvCount; i++)  {
					JSONObject lvCache = lvCA.getJSONObject(i);
					Log.v(TAG, "Jason says: "+lvCache);
					
					/*
					 * type:
					 * 2: Tradi
					 * 3: Multi
					 * 4: Virtual
					 * 5: Letterbox
					 * 6: Event
					 * 8: Mystery
					 * 9: Project API
					 * 11: Webcam
					 * 13: CITO
					 * 137: Earthcache
					 * 453: Megaevent
					 * 1858: WIGO
					 * 
					 * waypoint:
					 * 2000: Parking
					 * 2001: Trailhead
					 * 2002: Stage of a Multicache
					 * 2003: Question to answer
					 * 2004: Reference Point
					 * 2005: Final Location
					 */
					
					int lvType = lvCache.getInt("ctid");
	
					Double lvLat = lvCache.getDouble("lat");
					Double lvLon = lvCache.getDouble("lon");
					String lvName = lvCache.getString("nn");
					String lvGCID = lvCache.getString("gc");
					int lvID = Integer.valueOf(lvCache.getString("id"));
					String lvIsActive = lvCache.getString("ia");	  // "true" means active, else inactive			
					String lvIsFound = lvCache.getString("f");	      // "true" means found, else not yet found
					
					
					String lvFilterFound = lvFilter.get("found");
					lvFilterFound = lvFilterFound == null ? "" : lvFilterFound;
					if ((lvIsFound.equals("true") && lvFilterFound.contains("f")) ||
						(lvIsFound.equals("false") && lvFilterFound.contains("n"))) {	
						lvContainer.addRow(new Object[] {lvGCID, 
								lvName, 
								lvType, 
								lvLat,
								lvLon,
								lvID,
								lvIsActive,
								lvIsFound,
								null,
								null,
								null});
						lvGCLookup.put(lvGCID, lvName);
						
						if (aShowAddWP && DEBUG_WAYPOINT) {
							lvContainer.addRow(new Object[] {
									"WP"+lvGCID, 
									"Parking for "+lvName, 
									2000, 
									lvLat+0.005,
									lvLon+0.005,
									null,
									null,
									null,
									lvGCID,
									lvGCLookup.get(lvGCID),
									null});
						}
					}
									
				}						
			} catch (JSONException e) {
				throw new ConnectorException("Error parsing JSON-response from server");
			}
	
		} catch (ConnectorException anExc) {
			lvContainer.addRow(new String[] {null, null, null, null, null, null, null, null, null, null, anExc.getMessage()});
			Log.e(TAG, "Exception at readCachesByCoord", anExc);
			ivIsAlreadyInited = false;
		}
		return lvContainer;
	
	}
	


	/**
	 * Reads details for a cache displayed on the map. Needs the ID from the cache gotten via
	 * readCachesByCoord().
	 * The resulting table has this information as rows:
	 * 
	 * "owner"    - owner of the cache
	 * "diff"     - diff of cache
	 * "terr"     - terrain of cache
	 * "url"      - url for online viewing
	 * "size"     - size of the cachecontainer
	 * "messag"   - errormessage
	 * 
	 * Errors are given in the column "message". If it's not null this will be 
	 * interpreted as an error in the frontend.
	 * 
	 * @param aGCID
	 * @return
	 */
	public synchronized Cursor readCacheDetails(String anId) {
	
		MatrixCursor lvContainer = new MatrixCursor(new String[] {"owner", "diff", "terr", "url", "size", "message"});
	
		Log.v(TAG, "reading cachedetails");
	
		try {
			
			boolean lvRC = init();
	
			if (!lvRC) {
				lvContainer.addRow(new String[] {null, null, null, null, null, "Could not login!"});
				return lvContainer;	
			}
			
			if (!ivLastRequestFromMap) {
				String initRequest = initRequest("http://www.geocaching.com/map/default.aspx", false);
				mapUserToken = getPart(initRequest, "userToken = '", "';");
				ivLastRequestFromMap = true;
				
			}
			
			String lvBody = new StringBuilder()
			.append("{\"dto\":{\"data\":{\"c\":2,\"m\":\"\",\"d\":\"")
					.append(anId)
			.append("\"},\"ut\":\"").append(mapUserToken).append("\"}}").toString();
					
			// Url for the request. Again the coords here don't matter
			String lvUrlS ="http://"+GCMAPS_HOST+"/map/default.aspx/MapAction";
			URI lvUrl;
			try {
				lvUrl = new URI(lvUrlS);
			} catch (URISyntaxException e) {
				throw new ConnectorException("Error parsing Url "+lvUrlS);
			}
			
			HttpPost post = new HttpPost(lvUrl);
			addHeaders(post, URL_MAP_INIT, false );		
			post.setHeader("Content-Type", "application/json; charset=utf-8");
			post.setHeader("X-Requested-With", "XMLHttpRequest");
			post.setHeader("Accept", "application/json, text/javascript, */*; q=0.01");
			setPostBody(post, lvBody);
			HttpResponse lvResponse = executeHttpRequest(post);
	
			String lvResult = readResponse(lvResponse, false);
			
			// the JSON result
			try {
				//				String lvJson = getPart(lvResult.toString(), "<ExtraData><![CDATA[", "</ExtraData>");
				JSONObject lvObj = new JSONObject(lvResult);
				lvObj = new JSONObject((String)lvObj.get("d"));
			
			JSONObject lvCache = lvObj.getJSONObject("cs");
			Log.v(TAG, "Jason says: "+lvCache);
			
			String lvDiff = lvCache.getString("d");
			lvDiff = getPart(lvDiff, "alt=\"", " out");
			
			String lvTerr = lvCache.getString("t");
			lvTerr = getPart(lvTerr, "alt=\"", " out");
			
			String lvOwner = lvCache.getString("cb");
			String lvCUrl = lvCache.getString("curl");
			
			/* sizes: 	micro
			 * 			small
			 * 			regular
			 * 			large
			 * 			other
			 * 			not chosen
			 * 			virtual
			 */
			
			String lvSize = lvCache.getString("cz");
			lvSize = getPart(lvSize, "Size: ", "\"");
			
			lvContainer.addRow(new Object[] {lvOwner,
					lvDiff,
					lvTerr,
					lvCUrl,
					lvSize,
					null});
			} catch (JSONException e) {
				throw new ConnectorException("Error parsing JSON-response from server.");
			}
			
			return lvContainer;	
						
	
		} catch (ConnectorException anExc) {
			lvContainer.addRow(new String[] {null, null, null, null, null, anExc.getMessage()});
			Log.e(TAG, "Exception at readCacheDetails", anExc);
			ivIsAlreadyInited = false;
		}
	
		return lvContainer;
	
	}

	/**
	 * Reads just the spoilers for a given cache. Errors are returned in the column "message". 
	 * If it's not null this will be interpreted as an error in the frontend. The other column 
	 * returned is "number" which contains the number of downloaded spoilers.
	 * 
	 * @param aGCID          GCID of the cache to download spoilers for
	 * @param aSpoilerDir    spoiler-directory or null if spoiler download not necessary
	 * @return               result-cursor with gpx-string or message column
	 */
	public synchronized Cursor readSpoilers(String aGCID, String aSpoilerDir) {
		
		Log.v(TAG, "reading spoilers for "+aGCID);
		
		MatrixCursor lvContainer = new MatrixCursor(new String[] {"number", "message"});
					
		try {
			boolean lvRC = init();
	
			if (!lvRC) {
				lvContainer.addRow(new String[] {null, "Could not login!"});
				return lvContainer;	
			}
			
			ivLastRequestFromMap = false;
			
			String lvUrlGPX = URL_LISTING+aGCID;
			String lvRequest = initRequest(lvUrlGPX, false);
			
			HashMap<String, String> lvSpoilerUrls = new HashMap<String, String>();
	
			if (null != aSpoilerDir) {
				File lvSpoilDir = new File(aSpoilerDir);
				if (lvSpoilDir.exists() && lvSpoilDir.isDirectory()) {
					lvSpoilerUrls = extractSpoilersFromCache(lvRequest);
				} else {
					lvContainer.addRow(new String[] {null, "Can't use spoiler dir: "+lvSpoilDir});
					return lvContainer;	
				}
			} else {
				lvContainer.addRow(new String[] {null, "No spoiler-directory given!"});
				return lvContainer;	
			}
		
			int lvNumberSpoilers = downloadSpoilers(aGCID, aSpoilerDir, lvSpoilerUrls);
			lvContainer.addRow(new String[] {String.valueOf(lvNumberSpoilers), null});
			
		} catch (ConnectorException anExc) {
			lvContainer.addRow(new String[] {null, anExc.getMessage()});
			Log.e(TAG, "Exception at readGPX", anExc);
			Log.e(TAG, "Viewstate was: "+ivLastViewState);
			ivIsAlreadyInited = false;
		} 
		return lvContainer;
	}

	/**
	 * Read a GPX-file from the Server for a given GCID. Returns a Cursor to the Result. 
	 * The resulting table should have only one row containing the gpx-data as string in the
	 * column "gpx". Errors are given in the column "message". If it's not null this will be 
	 * interpreted as an error in the frontend.
	 * 
	 * @param aGCID          GCID of the cache to download GPX for
	 * @param aSpoilerDir    spoiler-directory or null if spoiler download not necessary
	 * @return               result-cursor with gpx-string or message column
	 */
	public synchronized Cursor readGPX(String aGCID, String aSpoilerDir) {
	
	
		Log.v(TAG, "reading GPX for "+aGCID);
	
		MatrixCursor lvContainer = new MatrixCursor(new String[] {"gpx", "message"});
					
		try {
			boolean lvRC = init();
	
			if (!lvRC) {
				lvContainer.addRow(new String[] {null, "Could not login!"});
				return lvContainer;	
			}
			
			ivLastRequestFromMap = false;
			
			String lvUrlGPX = URL_LISTING+aGCID;
			String lvRequest = initRequest(lvUrlGPX, false);
			
			HashMap<String, String> lvSpoilerUrls = new HashMap<String, String>();
	
			if (null != aSpoilerDir) {
				File lvSpoilDir = new File(aSpoilerDir);
				if (lvSpoilDir.exists() && lvSpoilDir.isDirectory()) {
					lvSpoilerUrls = extractSpoilersFromCache(lvRequest);
				} else {
					lvContainer.addRow(new String[] {null, "Can't use spoiler dir: "+lvSpoilDir});
					return lvContainer;	
				}
			}
		
			String lvBody = new StringBuilder()
			.append(ivLastViewState)
			.append("&ctl00$ContentBody$btnGPXDL=GPX+eXchange+File").toString();
			
			URI lvUrl;
			try {
				lvUrl = new URI(lvUrlGPX);
			} catch (URISyntaxException e) {
				throw new ConnectorException("Error parsing URL "+lvUrlGPX);
			}
			
			Log.v(TAG, "!"+lvUrl+"!");
			
			HttpPost post = new HttpPost(lvUrl);
			addHeaders(post, lvUrlGPX, true);		
			setPostBody(post,  lvBody);
			HttpResponse lvResponse = executeHttpRequest(post);
					
			String lvResult = readResponse(lvResponse, true, false);
			if (!lvResult.startsWith("<?xml")) {
				lvContainer.addRow(new String[] {null, "Could not download GPX, maybe you are not a premium member?"});
				return lvContainer;
			}
			Log.v(TAG, lvResult.substring(0, 100));
			lvContainer.addRow(new Object[]{lvResult, null});
			
			downloadSpoilers(aGCID, aSpoilerDir, lvSpoilerUrls);
			
			
			return lvContainer;
			
		} catch (ConnectorException anExc) {
			lvContainer.addRow(new String[] {null, anExc.getMessage()});
			Log.e(TAG, "Exception at readGPX", anExc);
			Log.e(TAG, "Viewstate was: "+ivLastViewState);
			ivIsAlreadyInited = false;
		} 
		return lvContainer;
		
	}

	/**
	 * Tries to write a Log to the Server
	 * 
	 * @param aGCID The GC-Code of the Cache for the Log
	 * @param aType Type of the Log, see below
	 * @param aText Text of the Log
	 * @param day 
	 * @param month
	 * @param year
	 * @return Cursor with Message
	 */
	public synchronized Cursor pushLog(String aGCID, String aType, String aText, String day, String month, String year) {
		//http://www.geocaching.com/seek/log.aspx?wp=GC20GY3
		
		/*
		 * POST /seek/log.aspx?wp=GC20GY3 HTTP/1.1
	       Cookie: userid=d80cc684-8cd5-43ec-a6e6-7910062ad4b3; ASP.NET_SessionId=3nvjbl45nriyoa45srsusy55;
		 */
		
		/*
		 * LogTypes:
		 * 2: FOUND: Found it
		 * 3: DNF: DNF
		 * 4: COMMENT: Write Note
		 * 7: SBA: SBA
		 * 45: NM: Maint
		 */
		
		MatrixCursor lvResult = new MatrixCursor(new String[] {"message"});
		try {
			
			
			String lvLogType = "2";
			if ("DNF".equals(aType)) {
				lvLogType = "3";
			} else if ("COMMENT".equals(aType)) {
				lvLogType = "4";
			} else if ("SBA".equals(aType)) {
				lvLogType = "7";
			} else if ("NM".equals(aType)) {
				lvLogType = "45";
			}
			
			
			boolean lvRC = init();
	
			if (!lvRC) {
				lvResult.addRow(new String[] {"Could not login!"});
				return lvResult;	
			}
	
			ivLastRequestFromMap = false;
			
			String lvUrlLog = URL_LOG+aGCID;
			String lvResponse = initRequest(lvUrlLog, false);
			
			// Check if the LogType is allowed:
			
			int idx1 = lvResponse.indexOf("ctl00_ContentBody_LogBookPanel1_ddLogType");
			int idx2 = lvResponse.indexOf("</select>");
			if (idx1 == -1 || idx2 == -1) {
				lvResult.addRow(new String[] {"Error parsing response from server"});	
				return lvResult;
				
			}
			String lvOptions = lvResponse.substring(lvResponse.indexOf("<option", idx1), idx2)
			;
			lvOptions = lvOptions
					.replaceAll("<option[^>]*value=\"", "")
					.replaceAll("\"[^<]*</option>", " ")
					.replaceAll("[\t\n\r]*", "")
					.trim();
			List<String> list = Arrays.asList(lvOptions.split(" "));
			if (!list.contains(lvLogType)) {
				lvResult.addRow(new String[] {"The chosen logtype is not available for this cache"});	
				return lvResult;
			}
		
	
			String lvBody = new StringBuilder()
			.append("__EVENTTARGET=&__EVENTARGUMENT=&__LASTFOCUS=&")
			.append(ivLastViewState).append("&ctl00%24ContentBody%24LogBookPanel1%24ddLogType=")
			.append(lvLogType) // 4
			.append("&ctl00%24ContentBody%24LogBookPanel1%24DateTimeLogged=")
//			.append(month).append("%2F").append(day).append("%2F").append(year) 
			.append("&ctl00%24ContentBody%24LogBookPanel1%24DateTimeLogged%24Month=")
			.append(month) 
			.append("&ctl00%24ContentBody%24LogBookPanel1%24DateTimeLogged%24Day=")
			.append(day) 
			.append("&ctl00%24ContentBody%24LogBookPanel1%24DateTimeLogged%24Year=")
			.append(year) 
			.append("&ctl00%24ContentBody%24LogBookPanel1%24uxLogInfo=")
			.append(URLEncoder.encode(aText)) // htmlized text
//			.append("&ctl00%24ContentBody%24LogBookPanel1%24lbTravelBugs=")
//			.append("-1") // -1
			.append("&ctl00%24ContentBody%24LogBookPanel1%24LogButton=Submit+Log+Entry")
			.append("&ctl00%24ContentBody%24uxVistOtherListingGC=&language=on&logcount=on&gpx=on").toString();
			
			URI lvUrl;
			try {
				lvUrl = new URI(URL_LOG+aGCID);
			} catch (URISyntaxException e) {
				throw new ConnectorException("Could not parse URL "+URL_LOG+aGCID);
			}
			
			HttpPost post = new HttpPost(lvUrl);
			addHeaders(post, URL_LOG+aGCID, true);		
			setPostBody(post, lvBody);
			HttpResponse lvHttpResponse2 = executeHttpRequest(post);
			String lvResponse2 = readResponse(lvHttpResponse2, true, false);
			
			if (!lvResponse2.contains("ctl00$ContentBody$LogBookPanel1$lnkbtnDelete")) {
				lvResult.addRow(new String[] {"Could not post log."});
				return lvResult;
			}
			
			lvResult.addRow(new String[] {"OK"});
		} catch (ConnectorException anExc) {
			lvResult.addRow(new String[] {anExc.getMessage()});
			Log.e(TAG, "Exception at pushLog", anExc);
			ivIsAlreadyInited = false;
		}
	
		return lvResult;
	}

	/**
	 * Tries to write Field Notes to the Server
	 * 
	 * @param aFile a File containing the Field Notes
	 * @return Cursor with Message
	 */
	public synchronized Cursor pushFieldNotes(String aFile) {
	
		MatrixCursor lvResult = new MatrixCursor(new String[] {"message"});
		try {
			
			BufferedReader f;
			try {
				f = new BufferedReader(new InputStreamReader(new FileInputStream(aFile), Charset.forName("UTF-8")));
			} catch (IllegalCharsetNameException e) {
				throw new ConnectorException("Internal error accessing charset UTF-8");
			} catch (UnsupportedCharsetException e) {
				throw new ConnectorException("Internal error accessing charset UTF-8");
			} catch (FileNotFoundException e) {
				throw new ConnectorException("Could not access temporary Field Note file "+aFile);
			}
			
			StringBuilder lvFieldNoteBuilder = new StringBuilder();
			try {
				String lvReadLine = f.readLine();
				while (lvReadLine != null) {
					lvFieldNoteBuilder.append(lvReadLine).append("\r\n");
					lvReadLine = f.readLine();
				}
			} catch (IOException e) {
				throw new ConnectorException("Error reading temporary Field Note file "+aFile);
			}
			
			boolean lvRC = init();
			
			if (!lvRC) {
				lvResult.addRow(new String[] {"Could not login!"});
				return lvResult;	
			}
			
			ivLastRequestFromMap = false;
			
			String lvUrlFieldNotes = URL_FIELD_NOTES;
			initRequest(lvUrlFieldNotes, false);
			
			String lvBoundary="--------------------"+Long.toString(System.currentTimeMillis(), 16);
	
			String lvBody = new StringBuilder()
				.append(getMultiPartSegment(lvBoundary, "__EVENTTARGET", "", null))
				.append(getMultiPartSegment(lvBoundary, "__EVENTARGUMENT", "", null))
				.append(getMultiPartSegment(lvBoundary, "__VIEWSTATE", URLDecoder.decode(ivLastViewState).substring("__VIEWSTATE=".length()), null))
				.append(getMultiPartSegment(lvBoundary, "ctl00$ContentBody$FieldNoteLoader", lvFieldNoteBuilder.toString(), "geocache_visits.txt"))
				.append(getMultiPartSegment(lvBoundary, "ctl00$ContentBody$btnUpload", "Upload Field Note", null))
				.append(getMultiPartSegment(lvBoundary, "language", "on", null))
				.append(getMultiPartSegment(lvBoundary, "logcount", "on", null))
				.append(getMultiPartSegment(lvBoundary, "gpx", "on", null))
				.append(getMultiPartSegment(lvBoundary, null, null, null))
				.toString();
				
			URI lvUrl;
			try {
				lvUrl = new URI(URL_FIELD_NOTES);
			} catch (URISyntaxException e) {
				throw new ConnectorException("Could not parse Url "+URL_FIELD_NOTES);
			}
	
			HttpPost post = new HttpPost (lvUrl);
			addHeaders(post, URL_FIELD_NOTES, true);		
			post.setHeader("Content-Type", "multipart/form-data; boundary="+lvBoundary);
			setPostBody(post, lvBody);
			
			HttpResponse lvHttpResponse = executeHttpRequest(post);
			
			String lvResponse2 = readResponse(lvHttpResponse, true, false);
			
			if (lvResponse2.contains("0 records were successfully uploaded")) {
				lvResult.addRow(new String[] {"No Field Notes were accepted by the server."});
				return lvResult;
			} else if ( !lvResponse2.contains("records were successfully uploaded")) {
				lvResult.addRow(new String[] {"Could not upload Field Notes."});
				return lvResult;
			}
			
			lvResult.addRow(new String[] {"OK"});
		} catch (ConnectorException anExc) {
			lvResult.addRow(new String[] {anExc.getMessage()});
			Log.e(TAG, "Exception at pushFieldNotes", anExc);
			ivIsAlreadyInited = false;
		}
		return lvResult;
	}
	
	/**
	 * 
	 * @param aSelectionArgs
	 * @return
	 */
	public Cursor downloadMultiple(String[] aSelectionArgs) {
		Log.v(TAG, "Received download multiple request with these parameters: ");
		for (String s : aSelectionArgs) {
			Log.v(TAG, " - "+s);
		} 
		MatrixCursor lvResult = new MatrixCursor(new String[] {"file", "message"});
		lvResult.addRow(new Object[]{"/sdcard/georg_data/demo.gpx", ""});
		return lvResult;
	}

	public Cursor syncCaches() {
		Log.v(TAG, "Received syncCaches-Request");
		MatrixCursor lvResult = new MatrixCursor(new String[] {"file", "message"});
		lvResult.addRow(new Object[]{"/sdcard/georg_data/demo.gpx", null});
		return lvResult;
	}

	/**
	 * Used by the Settings-Activity to reset Connector-state when configuration changes
	 */
	public void reset() {
		ivIsAlreadyInited = false;
	}

	/**
	 * Creates Session with groundspeak (viewstate, login) and saves cookies
	 * @throws ConnectorException 
	 */
	private boolean init() throws ConnectorException {
		if (ivIsAlreadyInited) {
			return true;
		}
		Log.v(TAG, "Reinitialization!");
				
		ivLastRequestFromMap = false;
		boolean lvRC = login();
		ivIsAlreadyInited = lvRC;
		
		return lvRC;	
	}
	
	/**
	 * Does a login to Groundspeak and saves session- and user-cookie
	 * @return did it work?
	 * @throws ConnectorException
	 */
	private boolean login() throws ConnectorException {
		Log.v(TAG, "login");
		
		// this is the first request, no cookies for the server
		initRequest(URL_LOGIN, true);

		StringBuilder lvBody = new StringBuilder();
		try {
			
//			lvBody.append(ivLastViewState)
//			.append("&ctl00%24SiteContent%24tbUsername=")
//			.append(URLEncoder.encode(getPreference(GCHttpConnectorSettings.SETTINGS_USER), "UTF-8"))
//			.append("&ctl00%24SiteContent%24tbPassword=")
//			.append(URLEncoder.encode(getPreference(GCHttpConnectorSettings.SETTINGS_PASSWORD), "UTF-8"))
//			.append("&ctl00%24SiteContent%24cbRememberMe=on&ctl00%24SiteContent%24btnSignIn=Login").toString();
			
			lvBody.append(ivLastViewState)
				.append("&ctl00%24ContentBody%24tbUsername=")
				.append(URLEncoder.encode(getPreference(GCHttpConnectorSettings.SETTINGS_USER), "UTF-8"))
				.append("&ctl00%24ContentBody%24tbPassword=")
				.append(URLEncoder.encode(getPreference(GCHttpConnectorSettings.SETTINGS_PASSWORD), "UTF-8"))
				.append("&ctl00%24ContentBody%24cbRememberMe=on&ctl00%24ContentBody%24btnSignIn=Login").toString();
		} catch (UnsupportedEncodingException e) {
			throw new ConnectorException("Internal error while accessing UTF-8 encoding.");
		}
		
		URI lvUrl;
		try {
			lvUrl = new URI(URL_LOGIN);
		} catch (URISyntaxException e) {
			throw new ConnectorException("Error parsing URL "+URL_LOGIN);
		}

		
		HttpPost post = new HttpPost(lvUrl);
		addHeaders(post, URL_LOGIN, false);
		setPostBody(post, lvBody.toString());

		HttpResponse lvResp = executeHttpRequest(post);
		
		String lvResult = readResponse(lvResp, true);
		int rc = lvResp.getStatusLine().getStatusCode();
		
		if (302 == rc) {
			//Handle the login-302 and get the data alright
			
			if (!saveCookies()) {
				return false;
			}
			
			HttpGet get = new HttpGet(lvUrl.toString());
			
			addHeaders(get, URL_LOGIN_GET, true);	
			lvResp = executeHttpRequest(get);
			lvResult = readResponse(lvResp, true);
		} else {
			if (!saveCookies()) {
				return false;
			}
		}
		
		setViewStateFromResponse(lvResult); 
		
		Header[] lvAllHeaders = lvResp.getAllHeaders();
		
		for (Header lvHeader : lvAllHeaders) {
			Log.v(TAG, "Header: "+lvHeader);
		}
		
		// Response plausible?
		if (!lvResult.contains("You are logged in as")) {
			return false;
		}
		
		Log.v(TAG, "User " + ivUser);
		Log.v(TAG, "Session " + ivSession);
		return true;				
		
	}
	
	
	/**
	 * Get the cookies from the Cookie-Store
	 * @return
	 */
	private boolean saveCookies() {
		List<Cookie> lvCookies = ivCookieStore.getCookies();
		if (null == lvCookies || lvCookies.size()==0) {
			Log.e(TAG, "Error at login: No cookies\n");
			return false;
		}
		// Save the cookies for later
		for (Cookie aCookie : lvCookies) {
			String lvCookie = aCookie.getName();
			if (lvCookie.equalsIgnoreCase("userid")) {
				ivUser = lvCookie+"="+aCookie.getValue()+";";
			} else if (lvCookie.equalsIgnoreCase("ASP.NET_SessionId")) {
				ivSession = lvCookie+"="+aCookie.getValue()+";";					
			}
		}
		return true;
	}
	
	/**
	 * Initialization of the ViewState from a given URL. This is mostly necessary to do a second
	 * (POST)-request on the base of an existing ViewState
	 * 
	 * @param anUrl
	 * @throws ConnectorException 
	 */
	private String initRequest(String anUrl, boolean aIsPristine) throws ConnectorException {

		if (null != ivCookieStore) {
			ivCookieStore.clear();
		}
		URI lvUrl;
		try {
			lvUrl = new URI(anUrl);
		} catch (URISyntaxException e) {
			throw new ConnectorException("Internal error parsing Url "+anUrl);
		}
		Log.v(TAG, "init request URL: "+anUrl);
		
		HttpGet get = new HttpGet(lvUrl);
		if (!aIsPristine) {
			addHeaders(get, anUrl, true);
		}
		HttpResponse lvResponse = executeHttpRequest(get);
		
		String lvResp = readResponse(lvResponse, false);
		setViewStateFromResponse(lvResp);	
		return lvResp;
	} 
	
	/**
	 * Adding HTTP-Headers to requests
	 * 
	 * @param aMessage
	 * @param aReferrer
	 * @param aCookies
	 */
	private void addHeaders(AbstractHttpMessage aMessage, String aReferrer, boolean aCookies) throws ConnectorException {
		
		aMessage.setHeader("User-Agent", "Mozilla/5.0 (Windows; U; Windows NT 5.1; de; rv:1.9.0.2) Gecko/2008091620 Firefox/3.0.2");
		aMessage.setHeader("Accept", "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8");
		aMessage.setHeader("Accept-Language", "de-de,de;q=0.8,en-us;q=0.5,en;q=0.3");
		aMessage.setHeader("Accept-Encoding", "gzip,deflate");
		aMessage.setHeader("Accept-Charset", "ISO-8859-1,utf-8;q=0.7,*;q=0.7");
		aMessage.setHeader("Keep-Alive", "300");
		aMessage.setHeader("Proxy-Connection", "keep-alive");
		aMessage.setHeader("delta", "true");
		aMessage.setHeader("Cache-Control", "no-cache, no-cache, no-cache");
		aMessage.setHeader("Referer", aReferrer);
		if (aCookies) {
			if (null == ivUser) {
				aMessage.setHeader("Cookie", ivSession);			
			} else {
				aMessage.setHeader("Cookie", ivUser+" "+ivSession);			
			}
		}
		aMessage.setHeader("Content-Type", "application/x-www-form-urlencoded; charset=UTF-8");
		
		
	}
		
	/**
	 * This method creates a MultiPartSegment as part of a MultiPart-Form-Request
	 * @param aBoundary Boundary-String for this MultiPart-Form-Request
	 * @param aName Name of the parameter, if null: write end-Boundary
	 * @param aValue Value of the Parameter
	 * @param aFileName optional filename or null. If not null, text/plain is used as Content-Type
	 * @return segement for use in POST-data
	 */
	private String getMultiPartSegment(String aBoundary, String aName, String aValue, String aFileName) {
		
		// End-Tag
		if (aName == null) {
			return new StringBuilder(MULTIPART_PREFIX)
					.append(aBoundary)
					.append(MULTIPART_PREFIX)
					.append("\r\n").toString();
		}
		
		// Part of a MultiPart-Container
		
		StringBuilder lvResult = new StringBuilder()
				.append(MULTIPART_PREFIX)
				.append(aBoundary)
				.append("\r\n")
				.append("Content-Disposition: form-data; name=\"").append(aName).append("\"");
		
		if (aFileName != null) {
			lvResult.append("; filename=\"").append(aFileName).append("\"\r\nContent-Type: text/plain");
		}
		
		return lvResult.append("\r\n\r\n")
				.append(aValue)
				.append("\r\n").toString();
		
	}

	/**
	 * Wrapper for setting the POST-Body of the current request and handling the
	 * Exception 
	 * @param aPost
	 * @param aBody
	 * @throws ConnectorException
	 */
	private void setPostBody(HttpPost aPost, String aBody) throws ConnectorException {
		try {
			aPost.setEntity(new StringEntity(aBody, "UTF-8"));
		} catch (UnsupportedEncodingException e) {
			Log.e(TAG, "UnsupportedEncoding", e);
			throw new ConnectorException("Unsupported Encoding");
		}
	}
	
	/**
	 * Gets a part of a string that's between a start-string and an end-string.
	 * @param aString
	 * @param aStart
	 * @param anEnd
	 * @return
	 */
	private String getPart(String aString, String aStart, String anEnd) {
		
		int i1 = aString.indexOf(aStart)+aStart.length();
		int i2 = aString.indexOf(anEnd, i1);
		
		if (i1 == aStart.length()-1 || i2 == -1) {
			return null;
		}
		return aString.substring(i1, i2);
		
	}
	
	/**
	 * Parses a response from the server for a new ViewState and saves it in this instance
	 * 
	 * @param aResp
	 * @throws UnsupportedEncodingException
	 * @throws ConnectorException
	 */
	private void setViewStateFromResponse(String aResp)  throws ConnectorException {
		StringBuilder lvResult = new StringBuilder();
		//Log.v(TAG, aResp);
		String lvCount = getPart(aResp, "__VIEWSTATEFIELDCOUNT\" value=\"", "\" />");
		int count=null == lvCount? 1:Integer.valueOf(lvCount);
		//Log.v(TAG, lvCount);
		if (count>1) { 
			lvResult.append("__VIEWSTATEFIELDCOUNT=").append(count).append("&");
		}
		//Log.v(TAG, lvResult.toString());
		for (int i=0; i<count; i++) { 
			String key = "__VIEWSTATE"+(i==0?"":i);
			
			String lvPart = getPart(aResp, key+"\" value=\"", "\" />");
			if (null == lvPart) {
				Log.v(TAG, "Didn't get a Viewstate!");
				throw new ConnectorException("Unexpected response without VIEWSTATE from server!");
			}
			try {
				lvResult.append(i!=0?"&":"").append(key).append("=").append(URLEncoder.encode(lvPart, "UTF-8"));
			} catch (UnsupportedEncodingException e) {
				throw new ConnectorException("Internal error accessing UTF-8-encoding.");
			}
			//Log.v(TAG, "VSResult: "+lvResult.toString());
			
		}
		
		ivLastViewState = lvResult.toString();
	}
	

	/**
	 * Reads the response from a http-call and returns it as a string
	 * 
	 * @param lvConnection
	 * @return
	 * @throws ConnectorException
	 */
	private String readResponse(HttpResponse lvResp, boolean aIgnoreRedir) throws ConnectorException {	
		return readResponse(lvResp, false, aIgnoreRedir);
	}
	
	/**
	 * Reads the response from a http-call and returns it as a string with the option to keep the CRs
	 * 
	 * @param lvConnection
	 * @return
	 * @throws ConnectorException
	 */
	private String readResponse(HttpResponse lvResp, boolean aKeepCR, boolean aIgnoreRedir) throws ConnectorException {
		
		int rc = lvResp.getStatusLine().getStatusCode();
		Log.v(TAG, "Response Code is: "+rc);
		if (!(rc==200 || (rc==302 && aIgnoreRedir))) {
			throw new ConnectorException("Server return redirect or error-page");
		}

		InputStream resultingInputStream = null;
		
		Header encoding = lvResp.getEntity().getContentEncoding();
		
		if (encoding != null && encoding.getValue().equalsIgnoreCase("gzip")) {
			try {
				resultingInputStream = new GZIPInputStream(lvResp.getEntity().getContent());
			} catch (IOException e) {
				throw new ConnectorException("IO-Error reading compressed response from server.");
			}
		} else if (encoding != null && encoding.getValue().equalsIgnoreCase("deflate")) {
			try {
				resultingInputStream = new InflaterInputStream(lvResp.getEntity().getContent(), new Inflater(true));
			} catch (IOException e) {
				throw new ConnectorException("IO-Error reading compressed response from server.");
			}
		} else {
			try {
				resultingInputStream = lvResp.getEntity().getContent();
			} catch (IOException e) {
				throw new ConnectorException("IO-Error reading response from server.");
				
			}
		}
		return readInputStream(resultingInputStream, aKeepCR);
		
		
	}
	
	/**
	 * Reads an Inputstream returns it's content as a string with the option to keep the CRs
	 * 
	 * @param stream
	 * @param aKeepCR
	 * @return
	 * @throws ConnectorException
	 */
	private String readInputStream(InputStream stream, boolean aKeepCR) throws ConnectorException {
		BufferedReader rd = new BufferedReader(new InputStreamReader(stream, Charset.forName("UTF-8")), 8192);
		
		StringBuilder lvResult = new StringBuilder();
		String line;
		try {
			while ((line = rd.readLine()) != null) {
				lvResult.append(line+(aKeepCR?"\n":""));
			}
		} catch (IOException e) {
			throw new ConnectorException("IO-Error reading response from server.");
		}
		return lvResult.toString();
	}
	
 	/**
 	 * Extracts the spoiler-URLs from a cachelisting and returns a HashMap with
 	 * the links to the spoilers.
 	 * @param lvRequest
 	 * @return HashMap with Spoiler-Name => Url
 	 */
 	private HashMap<String, String> extractSpoilersFromCache(String lvRequest) {
 		Pattern p = Pattern.compile("http://img.geocaching.com/cache[^\"]*\" rel=\"lightbox");
 		Matcher lvMatcher = p.matcher(lvRequest);
 		HashMap<String, String> lvResult = new HashMap<String, String>();
 		while (lvMatcher.find()) {
 			String found = lvMatcher.group();
 			String lvUrl = found.substring(0, found.indexOf("\""));
 			String lvOriginalName = "PIC";
 			
 			int lvOriginalNameS = lvRequest.indexOf("<span>", lvMatcher.end());
 			if (-1 != lvOriginalNameS) {
 				int lvOriginalNameE = lvRequest.indexOf("</span>", lvOriginalNameS+1);
 				if (-1 != lvOriginalNameE) {
 					lvOriginalName=lvRequest.substring(lvOriginalNameS+6, lvOriginalNameE);
 				}
 			}
 			String lvName = lvOriginalName.toLowerCase();
 			int idx = 1;
 			while(lvResult.containsKey(lvName)) {
 				lvName=lvOriginalName.toLowerCase()+(idx++);
 			}
 			
 			lvResult.put(lvName, lvUrl);
 		}

		return lvResult;
	}

	/**
	 * Helps convert a filter-parameter to a HashMap
	 * @param aFilter FilterString. Syntax: key1=value1&key2=value2&...
	 * @return a HashMap with the data
	 */
	private HashMap<String, String> parseCacheFilter(String aFilter) {
		HashMap<String, String> lvResult = new HashMap<String, String>();
		
		String[] lvSplit = aFilter.split("&");
		for (String aFilterPart : lvSplit) {
			String[] lvSplit2 = aFilterPart.split("=");
			lvResult.put(lvSplit2[0], lvSplit2[1]);
		}
		return lvResult;
	}
	
	/**
 	 * Download Spoiler-URLs from a HashMap to a given directory and prefixes the file
 	 * with the given GCID. Returns the number of downloaded spoilers or throws an exception.
 	 * @param aGCID
 	 * @param aSpoilerDir
 	 * @param lvSpoilerUrls
 	 * @return
 	 * @throws ConnectorException
 	 */
 	private int downloadSpoilers(String aGCID, String aSpoilerDir,
			HashMap<String, String> lvSpoilerUrls) throws ConnectorException {
		// is size 0 if aSpoilerDir is 0 or otherwise incorrect
		if (lvSpoilerUrls.size() > 0 ) {
			for (String name : lvSpoilerUrls.keySet()) {
				String lvSpoilerUrl = lvSpoilerUrls.get(name);
				String lvExt = lvSpoilerUrl.substring(lvSpoilerUrl.lastIndexOf("."));
				File f = new File(aSpoilerDir, aGCID+"."+name+lvExt);
				FileOutputStream fos;
				try {
					fos = new FileOutputStream(f);
				} catch (FileNotFoundException e) {
					throw new ConnectorException("Could not write to file "+f+": "+e.getMessage());
	
				}
				URL u;
				try {
					u = new URL(lvSpoilerUrl);
				} catch (MalformedURLException e) {
					throw new ConnectorException("Could not parse Spoiler-URL: "+lvSpoilerUrl);
				}
				InputStream is;
				try {
					is = u.openStream();
				} catch (IOException e) {
					throw new ConnectorException("Could not open connection for spoiler-picture "+lvSpoilerUrl);
				}
				
				byte[] b = new byte[4096];
				int read;
				try {
					while ((read = is.read(b)) != -1) {
						fos.write(b, 0, read);
					}
					fos.close();
					is.close();
				} catch (IOException e) {
					throw new ConnectorException("IO-Error download spoiler "+lvSpoilerUrl);
				}
			}
			
		}
		return lvSpoilerUrls.size();
	}

 	private String getPreference(String aWhich, String aDefault) {
 		SharedPreferences lvPref = ivCtx.getSharedPreferences(GCHttpConnectorSettings.SETTINGS_ID, Context.MODE_PRIVATE);
 		return lvPref.getString(aWhich, aDefault == null ? "" : aDefault);	
 	}
	/**
 	 * Get's preference-values for this application
 	 * @param aWhich
 	 * @return
 	 */
 	private String getPreference(String aWhich) {
 		return getPreference(aWhich, null);
	}

	/**
	 * Throws an Exception if no internet-connection is available
	 * @throws ConnectorException
	 */
	private void connectedOrThrow() throws ConnectorException {
		ConnectivityManager lvCM = (ConnectivityManager) ivCtx.getSystemService(Context.CONNECTIVITY_SERVICE);
		if (lvCM == null || lvCM.getActiveNetworkInfo() == null || !lvCM.getActiveNetworkInfo().isConnected()) {
			throw new ConnectorException("There is currently no internet connection!");
		}

	}


}
